#include <stdarg.h>

#include "devices.h"
#include "lib.h"
#include "network.h"

/* non-SPI pins */
#define CS	(1 << 27)
#define RESET	(1 << 28)

/* SPI registers */
#define SPI_BASE	0xe0020000
#define SPCR		(*(volatile unsigned char *) (SPI_BASE + 0x0))
#define SPSR		(*(volatile unsigned char *) (SPI_BASE + 0x4))
#define SPDR		(*(volatile unsigned char *) (SPI_BASE + 0x8))
#define SPCCR		(*(volatile unsigned char *) (SPI_BASE + 0xc))

/* SPCR flags */
#define MSTR	(1 << 5)

/* SPSR flags */
#define SPIF	(1 << 7)

/* commands for the enc28j60 */
#define RCR(x)	(0x00 | x)	/* read control register */
#define RBM	 0x3a		/* read buffer memory */
#define WCR(x)	(0x40 | x)	/* write control register */
#define WBM	 0x7a		/* write buffer memory */
#define BFS(x)	(0x80 | x)	/* bit field set */
#define BFC(x)	(0xa0 | x)	/* bit field clear */
#define SRC	 0xff		/* system reset command */

/* enc28j60 registers */
#define ERDPTL		0, 0x00	/* ethernet read pointer low */
#define ERDPTH		0, 0x01	/* ethernet read pointer high */
#define EWRPTL		0, 0x02	/* ethernet write pointer low */
#define EWRPTH		0, 0x03 /* ethernet write pointer high */
#define ETXSTL		0, 0x04	/* ethernet transmit buffer start low */
#define ETXSTH		0, 0x05	/* ethernet transmit buffer start high */
#define ETXNDL		0, 0x06	/* ethernet transmit buffer end low */
#define ETXNDH		0, 0x07	/* ethernet transmit buffer end high */
#define ERXSTL		0, 0x08	/* ethernet receive buffer start low */
#define ERXSTH		0, 0x09	/* ethernet receive buffer start high */
#define ERXNDL		0, 0x0a	/* ethernet receive buffer end low */
#define ERXNDH		0, 0x0b	/* ethernet receive buffer end high */
#define ERXRDPTL	0, 0x0c	/* ethernet receive data pointer low */
#define ERXRDPTH	0, 0x0d	/* ethernet receive data pointer high */
#define ERXWRPTL	0, 0x0e	/* ethernet receive write pointer low */
#define ERXWRPTH	0, 0xf	/* ethernet receive write pointer high */
#define ERXFCON		1, 0x18	/* ethernet receive filter control register */
#define EPKTCNT		1, 0x19	/* ethernet packet count */
#define MACON1		2, 0x00	/* MAC control register 1 */
#define MACON3		2, 0x02	/* MAC control register 3 */
#define MACON4		2, 0x03	/* MAC control register 4 */
#define MABBIPG		2, 0x04	/* MAC Back-To-Back Inter-Packet Gap */
#define MAIPGL		2, 0x06	/* MAC Non-Back-To-Back Inter-Packet Gap */
#define MAMXFLL		2, 0x0a	/* MAC maximum frame length low */
#define MAMXFLH		2, 0x0b	/* MAC maximum frame length high */
#define MIREGADR	2, 0x14	/* MII register address */
#define MIWRL		2, 0x16	/* MII write data low */
#define MIWRH		2, 0x17	/* MII write data high */
#define MAADR5		3, 0x00	/* MAC address part 5 */
#define MAADR6		3, 0x01	/* MAC address part 6 */
#define MAADR3		3, 0x02	/* MAC address part 3 */
#define MAADR4		3, 0x03	/* MAC address part 4 */
#define MAADR1		3, 0x04	/* MAC address part 1 */
#define MAADR2		3, 0x05	/* MAC address part 2 */
/* the following are in all benches */
#define ESTAT		0x1d	/* ethernet status register */
#define ECON2		0x1e	/* ethernet control register 2 */
#define ECON1		0x1f	/* ethernet control register 1 */

/* enc28j60 PHY registers */
#define PHCON1		0x00	/* PHY control register 1 */
#define PHCON2		0x10	/* PHY control register 2 */
#define PHLCON		0x14	/* PHY LED control register */

const unsigned char mac_addr[6] = { MAC_ADDR };

static const struct {
	unsigned short type;
	void (*handler)(void *pkg, const unsigned char *src);
} ether_types[] = {
	{ ETH_IP,  ip_handle  },
	{ ETH_ARP, arp_handle }
};

/* Ethernet is defined in IEEE 802.3. */

struct ether_header {
	unsigned char dst[6];
	unsigned char src[6];
	unsigned short type;
} __attribute__ ((packed));

static void spi_write(unsigned char data)
{
	int stat;
	
	stat = SPSR;
	SPDR = data;
	while (!SPSR) ;
}

static unsigned char spi_read()
{
	int stat;

	stat = SPSR;
	SPDR = 0xff;
	while (!SPSR) ;
	
	return SPDR;
}

/* read the specified register from the current bench */
static unsigned char rcr(int reg)
{
	unsigned char rv;

	IOCLR = CS;
	spi_write(RCR(reg));
	/* NOTE: on MAC registers, here would be a dummy byte */
	rv = spi_read();
	IOSET = CS;

	return rv;
}

/* set the specified register in the current bench */
static void wcr(int reg, unsigned char data)
{
	IOCLR = CS;
	spi_write(WCR(reg));
	spi_write(data);
	IOSET = CS;
}

/* set a bitfield on the specified register in the current bench */
static void bfs(int reg, unsigned char data)
{
	IOCLR = CS;
	spi_write(BFS(reg));
	spi_write(data);
	IOSET = CS;
}

/* change to specified bench if needed */
static void change_bench(unsigned char bench)
{
	static unsigned char current_bench = 0;
	unsigned char econ1;

	if (bench != current_bench) {
		econ1 = rcr(ECON1);
		econ1 &= 0xfc;
		econ1 |= bench;
		wcr(ECON1, econ1);
		current_bench = bench;
	}
}

/* set the specified register in the specified bench */
static void register_set(unsigned char bench, int reg, unsigned char data)
{
	change_bench(bench);
	wcr(reg, data);
}

/* get the specified register in the specified bench */
static unsigned char register_get(unsigned char bench, int reg)
{
	change_bench(bench);
	return rcr(reg);
}

/* set a register on the PHY */
static void phyreg_set(int reg, unsigned short data)
{
	register_set(MIREGADR, reg);
	register_set(MIWRL, data & 0xff);
	register_set(MIWRH, (data & 0xff00) >> 8);
}

void enc28j60_init()
{
	volatile int i = 5000000;

	/* set direction of CS and RESET pins and set them to high (they are
	 * active-low) */
	IODIR |= CS | RESET;
	IOSET = CS | RESET;

	/* set SPI interface into master mode */
	SPCR = MSTR;

	/* set SPI clock to 1/8 of CPU clock. This is the highest (fastest)
	 * possible value. The enc28j60 supports up to 20 MHz. With a PLL of 4
	 * and SPCCR set to 8, we only have ~ 7.4 MHz. This means, we can not
	 * use full 10 MBit/s speed with a PLL lesser than 6. */
	SPCCR = 8;

	/* reset enc28j60. maybe this is not urgently needed, but we want to
	 * ensure that the chip is in a defined state. */
	IOCLR = CS;
	spi_write(SRC);
	IOSET = CS;

	/* wait till clock of network chip is ready */
	while (!(rcr(ESTAT) & 1)) ;

	/* ERXST and ERXND define start and end of the receive buffer in the 8
	 * kB RAM of the enc28j60. The transmit buffer is everywhere where the
	 * receive buffer is not.
	 * We set the rx buffer to be between 0x800 and 0x2000. */
	register_set(ERXSTL, 0x00);
	register_set(ERXSTH, 0x08);
	register_set(ERXNDL, 0x00);
	register_set(ERXNDH, 0x20);

	/* we tell the receive filter, that packets should be only accepted if
	 * their ethernet CRC sum is correct AND either of the following
	 * conditions are met:
	 * - packet is a broadcast packet (dest MAC addr is ffffffffffff)
	 * - packet is a multicast packet (LSB set in dest MAC addr)
	 * - packet is a unicast packet for us (dest MAC addr is our MAC addr)
	 * Thanks to this filter possibilities, we do not need to filter that much
	 * ourself. */
	register_set(ERXFCON, 0xa3);

	/* Enable packet receiving, tell the MAC that control frames should be
	 * discarded after being processed by the MAC and enable RX/TX flow
	 * control (needed for full-duplex) */
	register_set(MACON1, 0xd);

	/* the MAC should automatically pad all frames to 60 bytes and add a CRC
	 * checksum. Additionally, too big packages should be discarded, and it
	 * should run in full-duplex mode. */
	register_set(MACON3, 0x33);

	/* tell the MAC to wait, when the medium is occupied while
	 * transmitting */
	register_set(MACON4, 0x40);

	/* set maximum frame length to 1518 bytes. I think that is a good
	 * value. */
	register_set(MAMXFLL, 0xee);
	register_set(MAMXFLH, 0x05);

	/* the recommended setting for full-duplex */
	register_set(MABBIPG, 0x15);
	register_set(MAIPGL,  0x12);

	/* The receive filter needs to know our MAC address. */
	register_set(MAADR1, mac_addr[0]);
	register_set(MAADR2, mac_addr[1]);
	register_set(MAADR3, mac_addr[2]);
	register_set(MAADR4, mac_addr[3]);
	register_set(MAADR5, mac_addr[4]);
	register_set(MAADR6, mac_addr[5]);

	/* tell PHY to operate in full-duplex mode, too */
	phyreg_set(PHCON1, 0x100);

	/* This configures the LEDs:
	 * - First LED should display link status
	 * - Second LED should display transmit and receive activity
	 * Of course there are much more options, but I like this most. */
	phyreg_set(PHLCON, 0x3472);

	/* tell the chip that read and write pointers should be automatically
	 * incremented after reading and writing. This way, we save *much*
	 * overhead. */
	wcr(ECON2, 0x80);

	/* set receive enable bit */
	bfs(ECON1, 1 << 2);

	/* wait some time to ensure the chip is really ready */
	while (i--) ;
}

void enc28j60_transmit(const unsigned char *dest_macaddr, unsigned short type,
		       int frag, ...)
{
	struct ether_header ether_header;
	unsigned int i;
	int pkgend, j, len = 0;
	va_list ap;
	unsigned int clen;
	const char *cpkg;

	/* wait till last packet is transmitted */
	while (rcr(ECON1) & (1 << 3)) ;

	/* build ethernet header */
	memcpy(ether_header.dst, dest_macaddr, 6);
	memcpy(ether_header.src, mac_addr, 6);
	ether_header.type = type;

	/* write the packet to the buffer.
	 * The packet contains:
	 * - packet control byte (in our case 0x00)
	 * - ethernet header
	 * - data payload (all the fragments we got) */
	register_set(EWRPTL, 0x00);
	register_set(EWRPTH, 0x00);
	IOCLR = CS;
	spi_write(WBM);
	spi_write(0x00);
	for (i = 0; i < sizeof(struct ether_header); i++)
		spi_write(((char *) &ether_header)[i]);
	va_start(ap, frag);
	for (j = 0; j < frag; j++) {
		cpkg = va_arg(ap, const char *);
		clen = va_arg(ap, unsigned int);
		for (i = 0; i < clen; i++)
			spi_write(cpkg[i]);
		len += clen;
	}
	va_end(ap);
	IOSET = CS;

	/* tell the chip, where we stored the data */
	register_set(ETXSTL, 0x00);
	register_set(ETXSTH, 0x00);
	pkgend = sizeof(struct ether_header) + len;	/* points to last byte */
	register_set(ETXNDL, pkgend & 0xff);
	register_set(ETXNDH, (pkgend & 0xff00) >> 8);

	/* tell the chip that we want to send */
	bfs(ECON1, 1 << 3);
}

static void process_packet(void *pkg)
{
	unsigned int i;
	const struct ether_header *ether_header = pkg;

	/* we do not need to pass the ethernet header to the handler */
	pkg += sizeof(struct ether_header);

	for (i = 0; i < ARRAY_SIZE(ether_types); i++)
		if (ether_types[i].type == ether_header->type)
			ether_types[i].handler(pkg, ether_header->src);
}

static void receive_packet(short len)
{
	/* our job is, to receive the packet part of a packet, and then
	 * process it. */

	char pkgdata[len];
	int i;

	for (i = 0; i < len; i++)
		pkgdata[i] = spi_read();
	IOSET = CS;

	process_packet(pkgdata);
}

void enc28j60_recvloop()
{
	short nextpkt;	/* pointer in enc28j60's memory to the next packet */
	short pkglen;

	/* first packet is always written to the start of the receive buffer */
	nextpkt = 0x800;

	while (1) {
		if (!register_get(EPKTCNT))
			continue;

		/* there is at least one packet to read */

		/* place read pointer to the packet start */
		register_set(ERDPTL, nextpkt & 0xff);
		register_set(ERDPTH, (nextpkt & 0xff00) >> 8);

		/* every packet is built up of:
		 * - pointer to the next packet (or to the place where the next
		 *   packet will be written)
		 * - length of the packet
		 * - lots of boring information bits, but we ignore them
		 * - ethernet header
		 * - data payload
		 * - padding
		 * - CRC sum */
		IOCLR = CS;
		spi_write(RBM);
		nextpkt = spi_read();
		nextpkt |= spi_read() << 8;
		pkglen = spi_read();
		pkglen |= spi_read() << 8;
		spi_read();
		spi_read();
		/* the eth packet itself is handled by another function */
		receive_packet(pkglen);

		/* the enc28j60 has a such called "ERXRDPT" register. The chip
		 * treats it as a "DO NOT GO BEYOND THIS POINT" sign while
		 * writing freshly received data. It is used to disallow
		 * overwriting data not being completely processed.
		 * After we processed the package, we can set it immediately
		 * before the next packet, to "tell" the chip that we do not
		 * need this old, already processed packet anymore. */
		register_set(ERXRDPTL, (nextpkt - 2) & 0xff);
		register_set(ERXRDPTH, ((nextpkt - 2) & 0xff00) >> 8);

		/* decrement EPKTCNT, since we have one packet lesser to
		 * process. */
		bfs(ECON2, 1 << 6);
	}
}

unsigned short enc28j60_get_recv_window()
{
	unsigned short rxwrpt, rxrdpt, rxnd, rxst;

	rxwrpt = register_get(ERXWRPTL) | (register_get(ERXWRPTH) << 8);
	rxrdpt = register_get(ERXRDPTL) | (register_get(ERXRDPTH) << 8);
	rxnd = register_get(ERXNDL) | (register_get(ERXNDH) << 8);
	rxst = register_get(ERXSTL) | (register_get(ERXSTH) << 8);

	if (rxwrpt > rxrdpt)
		return (rxnd - rxst) - (rxwrpt - rxrdpt);
	else if (rxwrpt == rxrdpt)
		return rxnd - rxst;
	return rxrdpt - rxwrpt - 1;
}
