#include <stdbool.h>

#include "errno.h"
#include "fs.h"
#include "lib.h"
#include "network.h"

/* FTP is defined in RFC 959. */

/* This struct is used as context information both for ftp and for ftpdata
 * sessions. For each client, they share one context. */
struct ftp_context {
	/* tcp context of ftpdata session and ftpcontrol session */
	struct tcp_context *datactx, *ctlctx;

	/* data to push on control session */
	const char *tosend;

	/* current working directory */
	char *cwd;

	/* after pushing the data, the connection will be closed. */
	bool closecon;

	struct inode fileino;
	int filepos;
};

/* the context of the client which will open the data connection next. This is
 * set when we tell a client our port information.
 * To identify ftp context this way is reliable in low-load situations. */
static struct ftp_context *next_context;

void ftpdata_connect(struct tcp_context *tctx)
{
	struct ftp_context *fctx;

	if (!next_context) {
		/* this connection is not expected. */
		tcp_disconnect(tctx);
		return;
	}

	/* link this tcp context with the ftp context */
	tctx->userdata = next_context;

	next_context = 0;

	fctx = tctx->userdata;

	/* link ftp context with this tcp context */
	fctx->datactx = tctx;
}

void ftpdata_close(const struct tcp_context *tctx)
{
	struct ftp_context *fctx = tctx->userdata;

	fctx->datactx = 0;
}

void ftpdata_push(struct tcp_context *tctx, void *data, unsigned short len)
{
	/* not needed yet */
	(void) tctx;
	(void) data;
	(void) len;
}

void ftp_close(const struct tcp_context *tctx)
{
	struct ftp_context *fctx = tctx->userdata;
	struct tcp_context *tmp;

	if (fctx->datactx) {
		tmp = fctx->datactx;
		ftpdata_close(tmp);
		tcp_disconnect(tmp);
	}

	free(fctx->cwd);

	free(fctx);
}

static void ftp_sendpkg(struct tcp_context *tctx)
{
	struct ftp_context *fctx = tctx->userdata;
	struct tcp_context *tmp;

	if (fctx->tosend) {
		printf("%s", fctx->tosend);
		tcp_transmit_str(tctx, fctx->tosend);
		fctx->tosend = 0;
		return;
	}

	if (fctx->closecon) {
		/* ftp QUIT procedure:
		 *
		 * step 1: close data connection
		 * step 2: send 221 (closing control connection)
		 * step 3: close control connection */

		if (fctx->datactx) {
			tmp = fctx->datactx;	/* need to save pointer copy
						   since ftpdata_close() sets it
						   to 0 */
			ftpdata_close(tmp);
			tcp_disconnect(tmp);
		}

		tcp_transmit_str(tctx, "221 Goodbye\r\n");

		ftp_close(tctx);
		tcp_disconnect(tctx);

		return;
	}

	tcp_stop_transmit(tctx);
}

void ftp_connect(struct tcp_context *tctx)
{
	struct ftp_context *fctx;
	tctx->userdata = malloc(sizeof(struct ftp_context));

	fctx = tctx->userdata;

	fctx->datactx = 0;
	fctx->ctlctx = tctx;
	fctx->closecon = 0;

	fctx->cwd = strdup("/");

	/* send 220 ("Service Ready") */
	fctx->tosend = "220 Service ready\r\n";
	tcp_start_transmit(tctx, ftp_sendpkg);
}

static void ftpdata_endtransmit(struct tcp_context *tctx)
{
	struct ftp_context *fctx = tctx->userdata;

	fctx->tosend = "226 Transmit finished\r\n";
	tcp_start_transmit(fctx->ctlctx, ftp_sendpkg);
	tcp_disconnect(tctx);
}

static void ftpdata_sendfile(struct tcp_context *tctx)
{
	struct ftp_context *fctx = tctx->userdata;
	char buf[1024];
	int i;

	i = minix_read_data(&fctx->fileino, fctx->filepos, buf, sizeof(buf));
	fctx->filepos += i;

	tcp_transmit_pkg(tctx, buf, i);

	if (i < (int) sizeof(buf))
		/* transmit finished (end of file reached) */
		ftpdata_endtransmit(tctx);
}

static void format_filemode(char *fm, unsigned short mode)
{
	int l, r;

	if (S_ISDIR(mode))
		fm[0] = 'd';
	else
		fm[0] = '-';

	/* we copy this string to fm and then set '-' were needed */

	memcpy(&fm[1], "rwxrwxrwx", 10);

	for (l = 1, r = 8; l < 10; l++, r--)
		if (!(mode & 1 << r))
			fm[l] = '-';
}

static void format_moddate(char *md, int mtime)
{
	static const char month[12][4] = { "Jan", "Feb", "Mar", "Apr",
					   "May", "Jun", "Jul", "Aug",
					   "Sep", "Oct", "Nov", "Dec" };
	struct tm ts;

	gmtime(mtime, &ts);

	/* TODO: according to POSIX, year should be printed if the file is older
	 * than 6 month. */

	sprintf(md, "%s %2i %2i:%02i",
		month[(int) ts.mon], ts.mday, ts.hour, ts.min);
}

static void ftpdata_senddir(struct tcp_context *tctx)
{
	struct ftp_context *fctx = tctx->userdata;
	char sendbuf[1400];
	struct dent readbuf[16]; /* send 16 file infos (1 block) at once */
	struct inode ino;
	int bts;
	int i, j = 0;
	int c;

	char filemode[11];	/* d, three times rwx, \0 --> 11 chars */
	char moddate[13];	/* Month (3), ' ', mday (2), '  ', year (4),
				   \0 --> 13 chars */

	bts = minix_read_data(&fctx->fileino, fctx->filepos, readbuf,
			      sizeof(readbuf));
	fctx->filepos += bts;

	for (i = 0; i < bts; i += sizeof(struct dent)) {
		c = i / sizeof(struct dent);
		minix_get_inode(readbuf[c].ino, &ino);
		format_filemode(filemode, ino.mode);
		format_moddate(moddate, ino.mtime);
		j += sprintf(&sendbuf[j], "%s %3i %5i %3i %10i %s %s\r\n",
			     filemode, ino.nlinks, ino.uid, ino.gid,
			     (int) ino.size, moddate, readbuf[c].name);
	}

	tcp_transmit_pkg(tctx, sendbuf, j);

	if (bts < (int) sizeof(readbuf))
		/* end of directory reached */
		ftpdata_endtransmit(tctx);
}

static int resolve_path(char *buf, const char *path, const char *cwd)
{
	int bytesleft = PATH_MAX;
	int i = 0, j;
	int lastslash;

	if (*path == '/') {
		/* If path begins with '/', searching starts at the root. */

		buf[i++] = '/';
		bytesleft--;
		path++;		/* do not copy '/' again later */
	} else {
		/* If path does not begin with '/', searching starts at the
		 * current working directory. */

		if (strlen(cwd) >= bytesleft)
			return -ENAMETOOLONG;

		strcpy(&buf[i], cwd);
		i += strlen(cwd);
		bytesleft -= strlen(cwd);
	}

	if (strlen(path) >= bytesleft)
		return -ENAMETOOLONG;

	strcpy(&buf[i], path);
	i += strlen(path);

	lastslash = 0;
	for (j = 0; j < i; j++) {
		if (buf[j] == '/' && buf[j + 1] == '.' && buf[j + 2] == '.'
		    && (buf[j + 3] == '/' || buf[j + 3] == 0)) {
			/* /dev/.. should resolve to / */
			strcpy(&buf[lastslash], &buf[j + 3]);
			i -= j - lastslash + 3;
		} else if (buf[j] == '/' && buf[j + 1] == '.'
			   && (buf[j + 3] == '/' || buf[j + 3] == 0)) {
			/* /dev/. should resolve to /dev/ */
			strcpy(&buf[j], &buf[j + 2]);
			i -= 2;
		} if (buf[j] == '/' && buf[j + 1] == '/') {
			/* /dev// should resolve to /dev/ */
			strcpy(&buf[j], &buf[j + 1]);
			i--;
			j--;	/* /dev/// should also resolve to /dev/ */
		}

		if (buf[j] == '/')
			lastslash = j;
	}

	buf[i] = 0;

	return 0;
}

/* tries to open file, sends error code when it fails.
 * returns positive inode number on success, 0 on failure. */
static int try_open(const char *filename, struct tcp_context *tctx)
{
	struct ftp_context *fctx = tctx->userdata;
	int rv;

	rv = minix_open(filename);

	if (rv < 0) {
		switch (-rv) {
		case ELOOP:
			fctx->tosend = "550 Too many levels of links\r\n";
			break;
		case EINVAL:
			fctx->tosend = "501 Invalid filename\r\n";
			break;
		case ENOTDIR:
			fctx->tosend = "550 Part of path not a directory\r\n";
			break;
		case ENOENT:
			fctx->tosend = "550 No such file or directory\r\n";
			break;
		default:
			fctx->tosend = "550 Does not work (dunno why)\r\n";
		};
		tcp_start_transmit(tctx, ftp_sendpkg);
		return 0;
	}

	return rv;
}

/* returns true on success */
static bool open_file(const char *filename, struct tcp_context *tctx)
{
	struct ftp_context *fctx = tctx->userdata;
	char buf[PATH_MAX+1];
	int ino;

	if (resolve_path(buf, filename, fctx->cwd) == -ENAMETOOLONG) {
		fctx->tosend = "500 Name too long\r\n";
		tcp_start_transmit(tctx, ftp_sendpkg);
		return 0;
	}

	fctx->filepos = 0;
	ino = try_open(buf, tctx);

	if (!ino)
		return 0;

	minix_get_inode(ino, &fctx->fileino);

	return 1;
}

/* returns true on success */
static bool prepare_send(const char *filename, struct tcp_context *tctx,
			 bool dir)
{
	struct ftp_context *fctx = tctx->userdata;

	if (!open_file(filename, tctx))
		return 0;

	if (dir && !S_ISDIR(fctx->fileino.mode)) {
		fctx->tosend = "550 Not a directory\r\n";
		tcp_start_transmit(tctx, ftp_sendpkg);
		return 0;
	} else if (!dir && S_ISDIR(fctx->fileino.mode)) {
		fctx->tosend = "550 Not a file\r\n";
		tcp_start_transmit(tctx, ftp_sendpkg);
		return 0;
	}

	if (!fctx->datactx) {
		fctx->tosend = "425 Data connection not open\r\n";
		tcp_start_transmit(tctx, ftp_sendpkg);
		return 0;
	}

	fctx->tosend = "125 Starting transmit\r\n";
	tcp_start_transmit(tctx, ftp_sendpkg);
	return 1;
}

static void retr(const char *filename, struct tcp_context *tctx)
{
	struct ftp_context *fctx = tctx->userdata;
	if (prepare_send(filename, tctx, 0))
		tcp_start_transmit(fctx->datactx, ftpdata_sendfile);
}

static void list(const char *dirname, struct tcp_context *tctx)
{
	struct ftp_context *fctx = tctx->userdata;
	if (prepare_send(dirname, tctx, 1))
		tcp_start_transmit(fctx->datactx, ftpdata_senddir);
}

static void size(const char *filename, struct tcp_context *tctx)
{
	struct ftp_context *fctx = tctx->userdata;
	static char answer[15];	/* "213", ' ', 10 bytes size, "\r\n\0" --> 17 */

	if (!open_file(filename, tctx))
		return;

	sprintf(answer, "213 %i\r\n", (int) fctx->fileino.size);

	fctx->tosend = answer;
	tcp_start_transmit(tctx, ftp_sendpkg);
}

static void pwd(struct tcp_context *tctx)
{
	struct ftp_context *fctx = tctx->userdata;
	static char answer[9+PATH_MAX];	/* "257 \"", PATH_MAX, "\"\r\n\0"
					   --> 9+PATH_MAX */

	sprintf(answer, "257 \"%s\"\r\n", fctx->cwd);

	fctx->tosend = answer;
	tcp_start_transmit(tctx, ftp_sendpkg);
}

static void cwd(const char *newcwd, struct tcp_context *tctx)
{
	struct ftp_context *fctx = tctx->userdata;
	char buf[PATH_MAX+2];
	int l;

	if (resolve_path(buf, newcwd, fctx->cwd) == -ENAMETOOLONG) {
		fctx->tosend = "500 Name too long\r\n";
		tcp_start_transmit(tctx, ftp_sendpkg);
		return;
	}

	/* cause its a directory, ensure it has a trailing slash */
	l = strlen(buf);
	if (buf[l-1] != '/') {
		buf[l] = '/';
		buf[l+1] = 0;
	}

	/* check whether dir is openable */
	if (!try_open(buf, tctx))
		return;

	free(fctx->cwd);
	fctx->cwd = strdup(buf);

	fctx->tosend = "250 Changed directory\r\n";
	tcp_start_transmit(tctx, ftp_sendpkg);
}

/* makes the string to uppercase till either '\0' or ' ' comes */
static void strtoupper(char *str)
{
	while (*str != ' ' && *str != 0) {
		if ('a' <= *str && *str <= 'z')
			*str += 'A' - 'a';
		str++;
	}
}

void ftp_push(struct tcp_context *tctx, void *pkg, unsigned short pkglen)
{
	struct ftp_context *fctx = tctx->userdata;
	char *cb = pkg;

	if (pkglen < 2) {
		fctx->tosend = "500 Invalid syntax\r\n";
		tcp_start_transmit(tctx, ftp_sendpkg);
	}

	cb[pkglen - 2] = 0;

	strtoupper(cb);

	if (!strcmp("QUIT", cb)) {
		fctx->closecon = 1;
		tcp_start_transmit(tctx, ftp_sendpkg);
	} else if (!strcmp("SYST", cb)) {
		/* send system type. Possible types are listed at:
		 * http://www.iana.org/assignments/operating-system-names
		 * Since we act like an UNIX ftp server, it's best to say UNIX
		 * here, though we are not UNIX. */
		fctx->tosend = "215 UNIX like system\r\n";
		tcp_start_transmit(tctx, ftp_sendpkg);
	} else if (!strncmp("USER ", cb, 5)) {
		fctx->tosend = "230 Logged in\r\n";
		tcp_start_transmit(tctx, ftp_sendpkg);
	} else if (!strcmp("PASV", cb)) {
		/* entering passive mode */
		fctx->tosend = "227 Listening on " STR(IP_ADDR) ",0,20\r\n";
		next_context = fctx;
		tcp_start_transmit(tctx, ftp_sendpkg);
	} else if (!strncmp("PORT ", cb, 5)) {
		fctx->tosend = "425 Active mode not supported\r\n";
		tcp_start_transmit(tctx, ftp_sendpkg);
	} else if (!strncmp("TYPE ", cb, 5)) {
		/* we do not distinguish between file types. So it's ok */
		fctx->tosend = "200 OK\r\n";
		tcp_start_transmit(tctx, ftp_sendpkg);
	} else if (!strncmp("STRU ", cb, 5)) {
		/* just tell its okay... */
		fctx->tosend = "200 OK\r\n";
		tcp_start_transmit(tctx, ftp_sendpkg);
	} else if (!strncmp("MODE ", cb, 5)) {
		if (cb[5] == 'S')
			/* stream mode - supported and our default */
			fctx->tosend = "200 OK\r\n";
		else
			fctx->tosend = "504 Unsupported mode\r\n";
		tcp_start_transmit(tctx, ftp_sendpkg);
	} else if (!strcmp("NOOP", cb)) {
		fctx->tosend = "200 OK\r\n";
		tcp_start_transmit(tctx, ftp_sendpkg);
	} else if (!strncmp("SIZE ", cb, 5)) {
		size(&cb[5], tctx);
	} else if (!strcmp("PWD", cb)) {
		pwd(tctx);
	} else if (!strncmp("RETR ", cb, 5)) {
		retr(&cb[5], tctx);
	} else if (!strcmp("LIST", cb)) {
		list(".", tctx);
	} else if (!strncmp("LIST ", cb, 5)) {
		list(&cb[5], tctx);
	} else if (!strncmp("CWD ", cb, 4)) {
		cwd(&cb[4], tctx);
	} else if (!strcmp("CDUP", cb)) {
		cwd("..", tctx);
	} else {
		printf("ftp: got unknown command \"%s\"\n", cb);
		fctx->tosend = "502 Unknown command\r\n";
		tcp_start_transmit(tctx, ftp_sendpkg);
	}
}
