#include "errno.h"
#include "fs.h"
#include "lib.h"
#include "network.h"

/* HTTP is defined in RFC 2616. */

/* mime types. To extend this table, look in /etc/mime.types. Most common mime
 * types on the system can be found with:
 *  find / | sed -e 's/^.*\.//' | sort | uniq -c | sort -g
 * Maybe this list contains types common for home directories of software
 * development machines instead of types common for webservers. */
static const struct {
	const char *ending;
	const char *type;
} mime_types[] = {
	{ ".asc",	"text/plain"		},
	{ ".c"	,	"text/x-csrc"		},
	{ ".cls",	"text/x-tex"		},
	{ ".diff",	"text/plain"		},
	{ ".h",		"text/x-chdr"		},
	{ ".ltx",	"text/x-tex"		},
	{ ".o",		"application/x-object"	},
	{ ".pdf",	"application/pdf"	},
	{ ".png",	"image/png"		},
	{ ".pot",	"text/plain"		},
	{ ".sh",	"text/x-sh"		},
	{ ".sty",	"text/x-tex"		},
	{ ".tex",	"text/x-tex"		},
	{ ".text",	"text/plain"		},
	{ ".txt",	"text/plain"		},
	{ ".xml",	"application/xml"	},
	{ ".xpm",	"image/x-xpixmap"	},
	{ ".xsl",	"application/xml"	}
};

struct http_ctx {
	enum {
		header,	/* next time the header should be sent		*/
		data,	/* next time sending data should be continued	*/
		close	/* next time the connection should be closed	*/
	} to_send;

	int status_code;
	const char *status_desc;

	const char *content_type;

	struct inode fileino;
	int fileaddr;
};

static const char *find_mime_type(const char *fn)
{
	unsigned int i;
	int flen;
	int tlen;

	flen = strlen(fn);

	for (i = 0; i < ARRAY_SIZE(mime_types); i++) {
		tlen = strlen(mime_types[i].ending);

		if (!strcmp(mime_types[i].ending, &fn[flen-tlen]))
			return mime_types[i].type;
	}

	/* no mime type found... just guess it's text/plain */
	return "text/plain";
}

void http_connect(struct tcp_context *ctx)
{
	ctx->userdata = malloc(sizeof(struct http_ctx));
}

void http_close(const struct tcp_context *ctx)
{
	free(ctx->userdata);
}

static int make_error_page(struct http_ctx *htctx, char *buf)
{
	return sprintf(buf,
		       "<?xml version=\"1.0\"?>"
		       "<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.1//EN\" "
		       "\"http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd\">"
		       "<html xmlns=\"http://www.w3.org/1999/xhtml\">"
		       "<head><title>%i - %s</title></head>"
		       "<body><h1>%i - %s</h1></body></html>",
			htctx->status_code, htctx->status_desc,
			htctx->status_code, htctx->status_desc);
}

static void http_send_packet(struct tcp_context *ctx)
{
	struct http_ctx *htctx = ctx->userdata;
	char buf[1024];
	unsigned int len = 0, i;

	switch (htctx->to_send) {
	case header:
		if (htctx->status_code == 200) {
			len = htctx->fileino.size;

			/* if everything is ok, content type is already set in
			 * http_push() */

			/* do not allow directory index */
			if (S_ISDIR(htctx->fileino.mode)) {
				htctx->status_code = 403;
				htctx->status_desc = "Forbidden";
			}
		} if (htctx->status_code != 200) {
			len = make_error_page(htctx, buf);
			htctx->content_type = "application/xhtml+xml";
		}

		i = sprintf(buf,
			    "HTTP/1.1 %i %s\r\n"
			    "Content-Type: %s\r\n"
			    "Content-Length: %i\r\n"
			    "\r\n",
			    htctx->status_code, htctx->status_desc,
			    htctx->content_type, len);
		tcp_transmit_pkg(ctx, buf, i);
		htctx->to_send = data;
		return;
	case data:
		if (htctx->status_code != 200) {
			i = make_error_page(htctx, buf);
			tcp_transmit_pkg(ctx, buf, i);
			htctx->to_send = close;
			return;
		}

		i = minix_read_data(&htctx->fileino, htctx->fileaddr, buf,
				    sizeof(buf));
		tcp_transmit_pkg(ctx, buf, i);
		if (i < sizeof(buf))
			htctx->to_send = close;
		else
			htctx->fileaddr += i;
		return;
	case close:
		http_close(ctx);
		tcp_disconnect(ctx);
		return;
	};
}

void http_push(struct tcp_context *ctx, void *pkg, unsigned short pkglen)
{
	char *cb = pkg;
	char *tmp;
	struct http_ctx *htctx = ctx->userdata;
	int ino;

	cb[pkglen - 1] = 0;

	if (!strncmp(cb, "GET ", 4)) {
		htctx->to_send = header;

		tmp = strchr(&cb[4], ' ');
		if (!tmp) {
			htctx->status_code = 400;
			htctx->status_desc = "Bad Request";
			goto get_start;
		}
		*tmp = 0;

		ino = minix_open(&cb[4]);

		if (ino < 0) {
			switch (-ino) {
			case EINVAL:
				htctx->status_code = 400;
				htctx->status_desc = "Bad Request";
				break;
			case ELOOP:
			case ENOTDIR:
			case ENOENT:
				htctx->status_code = 404;
				htctx->status_desc = "Not Found";
				break;
			default:
				htctx->status_code = 500;
				htctx->status_desc = "Internal Server Error";
			};
		} else {
			minix_get_inode(ino, &htctx->fileino);
			htctx->status_code = 200;
			htctx->status_desc = "OK";
			htctx->fileaddr = 0;
			htctx->content_type = find_mime_type(&cb[4]);
		}

get_start:
		tcp_start_transmit(ctx, http_send_packet);
		return;
	}
}
