#include <stdarg.h>
#include <stdbool.h>

#include "devices.h"
#include "lib.h"
#include "network.h"

/*
 * stdio
 */

/* printf' buffer size */
#define PRINTF_BUF	64

/* makros for printf */
#define CHAR_CPY(c)				\
	dest[rv++] = c;
#define STR_CPY(s)				\
	for (i = minlen - strlen(s); i > 0; i--)\
		CHAR_CPY(fillc)			\
	i = 0;					\
	while (s[i])				\
		CHAR_CPY(s[i++]);

static void itos(unsigned long id, char *s, int base, bool is_signed)
{
	static const char chars[] = "0123456789abcdef";
	unsigned char j = 0;
	bool negative = 0;
	char *p1, *p2, tmp;

	if (base < 2 || base > 16)
		base = 10;

	if (is_signed && (signed) id < 0) {
		negative = 1;
		id = -(signed) id;
	}

	do {
		s[j++] = chars[id % base];
		id /= base;
	} while (id != 0);

	if (negative)
		s[j++] = '-';

	s[j] = 0;

	/* reverse the buf (we have written in wrong order) */
	p1 = s;
	p2 = &s[j - 1];
	while (p1 < p2) {
		tmp = *p1;
		*p1 = *p2;
		*p2 = tmp;
		p1++;
		p2--;
	}
}

int vsprintf(char *dest, const char *format, va_list ap)
{
	int rv = 0, i;
	char c;
	char buf[12];
	const char *s;

	int minlen;
	char fillc;

	while ((c = *format++) != 0) {
		/* variable references begin with '%' */
		if (c != '%') {
			CHAR_CPY(c);
			continue;
		}

		minlen = 0;
		fillc = ' ';

		c = *format++;

		/* may be followed by '0' to tell that they should be
		 * right-aligned by filling with '0' instead of ' ' */
		if (c == '0') {
			fillc = c;
			c = *format++;
		}

		/* may be followed by the minimum field size */
		while ('0' <= c && c <= '9') {
			minlen *= 10;
			minlen += c - '0';
			c = *format++;
		}

		/* must be followed by type indicator */
		switch (c) {
			case 'i':
				itos(va_arg(ap, int), buf, 10, 1);
				STR_CPY(buf);
				break;
			case 'x':
				itos(va_arg(ap, int), buf, 16, 0);
				STR_CPY(buf);
				break;
			case 's':
				s = va_arg(ap, const char *);
				if (!s)
					s = "(null)";
				STR_CPY(s);
				break;
			default:
				CHAR_CPY(c);
		}
	}

	dest[rv] = 0;

	return rv;
}

int sprintf(char *buf, const char *fmt, ...)
{
	int rv;
	va_list ap;

	va_start(ap, fmt);
	rv = vsprintf(buf, fmt, ap);
	va_end(ap);

	return rv;
}

int printf(const char *fmt, ...)
{
	char buf[PRINTF_BUF];
	int rv, i;
	va_list ap;

	va_start(ap, fmt);
	rv = vsprintf(buf, fmt, ap);
	va_end(ap);

	for (i = 0; i < rv; i++)
		udpcons_putchar(buf[i]);

	return rv;
}

int puts(const char *s)
{
	while (*s)
		udpcons_putchar(*s++);
	udpcons_putchar('\n');
	return 0;
}

/*
 * string
 */

void *memcpy(void *dest, const void *src, int n)
{
	char *d = dest;
	const char *s = src;

	while (n--)
		*d++ = *s++;

	return dest;
}

int memcmp(const void *s1, const void *s2, int n)
{
	const unsigned char *c1 = s1, *c2 = s2;
	int d = 0;

	while (n--) {
		d = (int) *c1++ - (int) *c2++;
		if (d)
			break;
	}

	return d;
}

char *strchr(const char *s, int c)
{
	while (*s != (char) c) {
		if (!*s)
			return 0;
		s++;
	}

	return (char *) s;
}

int strlen(const char *s)
{
	int i = 0;

	while (s[i++] != 0) ;

	return i - 1;
}

int strncmp(const char *s1, const char *s2, int n)
{
	int c = 0;

	while (n-- && !(c = *s1++ - *s2++) && (*s1 || *s2)) ;

	return c;
}

int strcmp(const char *s1, const char *s2)
{
	int c = 0;

	while (!(c = *s1++ - *s2++) && (*s1 || *s2)) ;

	return c;
}

char *strdup(const char *s)
{
	int len = strlen(s) + 1;	/* +1 because of \0 */
	char *rv = malloc(len);

	return memcpy(rv, s, len);
}

char *strcpy(char *d, const char *s)
{
	return memcpy(d, s, strlen(s)+1);
}

/*
 * netinet/in
 */

unsigned short htons(unsigned short b)
{
	return (b << 8) | (b >> 8);
}

unsigned short ntohs(unsigned short b)
{
	return htons(b);
}

unsigned int htonl(unsigned int b)
{
	return (b >> 24) | (b << 24) | ((b & 0x00FF0000) >> 8)
		| ((b & 0x0000FF00) << 8);
}

unsigned int ntohl(unsigned int b)
{
	return htonl(b);
}

/*
 * stdlib
 */

#define BLOCKS		3072	/* 48 kB */
#define BLOCKSIZE	16
static struct {
	int bitmap[BLOCKS/32];	/* bitmap indicating which blocks are used */
	char blocks[BLOCKS*BLOCKSIZE];
} dm;

void *malloc(int size)
{
	int i, j;
	int blk;
	void *rv;

	size += 4;	/* needing an int to store length of allocation */

	blk = blocks(size, BLOCKSIZE);

	/* scan for enough free blocks */
	for (i = 0; i < BLOCKS; i++) {
		if (dm.bitmap[i / 32] & 1 << (i % 32))
			/* block used */
			continue;

		/* check whether there are enough continious free blocks */
		for (j = i + 1; j < i + blk && j < BLOCKS; j++)
			if (dm.bitmap[j / 32] & 1 << (j % 32))
				break;

		if (j < i + blk) {
			/* not enough free blocks */
			i = j;
			continue;
		}

		break;
	}

	if (i == BLOCKS)
		panic("Out of Memory.");

	/* mark blocks as used */
	for (j = i; j < blk + i; j++)
		dm.bitmap[j / 32] |= 1 << (j % 32);

	/* get address of freshly allocated allocation */
	rv = dm.blocks + i * BLOCKSIZE;

	/* save number of blocks */
	*(int *) rv = blk;

	/* return */
	return rv + 4;
}

void free(const void *ptr)
{
	int blknr, blkcnt;
	int i;

	if (!ptr)
		return;

	ptr -= 4;

	blkcnt = *(int *) ptr;
	blknr = ((unsigned long) ptr - (unsigned long) dm.blocks) / BLOCKSIZE;

	/* mark blocks as free */
	for (i = blknr; i < blknr + blkcnt; i++)
		dm.bitmap[i / 32] &= ~(1 << (i % 32));
}

/*
 * time
 */

/* days of a normal year */
static const char mdays_nonleap[12] = {
	31, 28, 31, 30, 31, 30,
	31, 31, 30, 31, 30, 31
};
#define YDAYS_NONLEAP	365

/* days of a leap year */
static const char mdays_leap[12] = {
	31, 29, 31, 30, 31, 30,
	31, 31, 30, 31, 30, 31
};
#define YDAYS_LEAP	366

static bool year_is_leap(int y)
{
	return ((y % 4 == 0) && (y % 100 != 0)) || (y % 400 == 0);
}

void gmtime(int tl, struct tm *ts)
{
	int i;
	int sub;
	const char *mdays;

	/* first subtract years */
	for (i = 70; ; i++) {
		if (year_is_leap(i + 1900))
			sub = YDAYS_LEAP * 24 * 60 * 60;
		else
			sub = YDAYS_NONLEAP * 24 * 60 * 60;

		if (tl - sub < 0)
			break;

		tl -= sub;
	}
	ts->year = i;

	/* then months of the year */
	mdays = year_is_leap(ts->year + 1900) ? mdays_leap : mdays_nonleap;
	for (i = 0; ; i++) {
		sub = mdays[i] * 24 * 60 * 60;
		if (tl - sub < 0)
			break;
		tl -= sub;
	}
	ts->mon = i;

	/* then days, hours, minutes, seconds */
	ts->mday = tl / (24 * 60 * 60) + 1; tl = tl % (24 * 60 * 60);
	ts->hour = tl / (60 * 60); tl = tl % (60 * 60);
	ts->min = tl / 60; tl = tl % 60;
	ts->sec = tl;
}

/*
 * nowhere
 */

unsigned short create_checksum(const void *buf, int size)
{
	int i;
	unsigned int rv = 0;
	const unsigned short *sb = buf;
	const unsigned char  *cb = buf;

	for (i = 0; i < size / 2; i++)
		rv += sb[i];

	if (size % 2 != 0)	/* odd number */
		rv += cb[size - 1];

	while (rv & 0xffff0000)
		rv = ((rv & 0xffff0000) >> 16) + (rv & 0x0000ffff);

	return ~rv;
}

void panic(const char *reason)
{
	printf("panic: %s - system halted.\n", reason);
	for (;;) ;
}

unsigned int blocks(int size, int blocksize)
{
	int r = size / blocksize;

	if (size % blocksize != 0)
		r++;

	return r;
}
