#include <stdbool.h>

#include "errno.h"
#include "devices.h"
#include "fs.h"
#include "lib.h"

/* The minix filesystem is build up of:
 *
 * 1 kb nothing (reserverd for bootloader etc.)
 * 1 kb superblock
 * (imap_blocks * BLOCKSIZE) bytes inode map
 * (zmap_blocks * BLOCKSIZE) bytes zone map
 * ninodes * sizeof(struct inode) bytes inode table
 * up to (ZONESIZE - 1) bytes nothing (to fill the current zone)
 * (nzones - firstdatazone) * ZONESIZE bytes data zones */

/* Some words about the inodes, zones and blocks:
 *
 * - The 0. inode is not used. It has its entry (set to 1) in the inode bitmap,
 *   but it has NO entry in the inode table.
 * - The root directory has always inode #1
 * - inode::zone uses the first 7 entries to store the numbers of the first 7
 *   zones in which the file is (direct zone). The 8. entry (inode::zone[7])
 *   stores the number of a zone containing this array for the next zones of the
 *   file (indirect zone). The 9. entry (inode::zone[8]) stores the zone number
 *   of a zone containing an array of zones like zone[7] (double indirect zone)
 *
 * - A block has a size of BLOCKSIZE bytes.
 *
 * - A zone has a size of ZONESIZE bytes.
 * - The WHOLE filesystem is seperated into zones. Not only the data area.
 *   superblock::firstdatazone points to the first zone usable for data (The
 *   zone after the last zone used for the inode table)
 * - The first unusable zones (the first superblock::firstdatazone zones) have
 *   NO entry in the zone bitmap.
 *
 * - Both zone and inode maps are bitmaps indicating whether the zone/inode is
 *   used or not. 1 = used, 0 = unused.
 * - To check, whether inode 'i' is used, do:
 *   (map[i / 8] & 1 << (i % 8)).
 * - To check, whether zone 'z' is used, do:
 *   z -= superblock::firstdatazone;
 *   (map[z / 8] & 1 << (z % 8)), because the unusable zones have no entry in
 *   the zone bitmap. */

/* Directories:
 *
 * In the data area of directory inodes is an array of dirents of the
 * directories/files stored in the directory. */

struct superblock {
	unsigned short ninodes;		/* usable inodes		*/
	unsigned short nzones;		/* total number of zones	*/
	unsigned short imap_blocks;	/* size of inode map [blocks]	*/
	unsigned short zmap_blocks;	/* size of zone map  [blocks]	*/
	unsigned short firstdatazone;	/* number of first data zone	*/
	unsigned short log_zone_size;	/* log2 of blocks/zone		*/
	unsigned long max_size;		/* maximum file size		*/
	unsigned short magic;		/* magic number			*/
	unsigned short state;		/* file system state		*/
};

/* some structs etc. are in fs.h */

/* magic for minix V1 filesystem with 30 character long filenames */
#define V1_MAGIC	0x138f

/* valid values for superblock::state */
#define STATE_CLEAN	1
#define STATE_DIRTY	2

#define BLOCKSIZE	1024

#define ZONESIZE	(BLOCKSIZE << sb.log_zone_size)

/* maximal number of links needed to find the file. This is
 * implementation-defined and we use it to detect endless loops on filesystems
 *
 * e.g. if A is a link to B, B is a link to C and C is a link to A, we would
 * recurse (really RECURSE, not iterate) in an endless loop when trying to
 * access A, B or C.
 *
 * After MAX_LINKS are reached, we return ELOOP. */
#define MAX_LINKS	10

/* start of the filesystem */
static unsigned int offset;

#define READ(addr, count, buf) \
	mmc_read(addr + offset, count, buf)

/* our superblock information */
static struct superblock sb;

/* true, if mounted read-only */
static bool ro;

/* Do the minix filesystem driver initalizations.
 * off is the start of the minix filesystem to be used. */
void minix_mount(unsigned int off)
{
	offset = off;

	/* read superblock. It starts at BLOCKSIZE cause the first block is
	 * unused. */
	READ(BLOCKSIZE, sizeof(struct superblock), &sb);

	if (sb.magic != V1_MAGIC)
		panic("Not a supported minix filesystem.");

	if (sb.state == STATE_DIRTY) {
		printf("Filesystem dirty -> mounted read-only.\n");
		ro = true;
	} else if (sb.state != STATE_CLEAN)
		panic("Invalid filesystem state.");
}

void minix_get_inode(int nr, struct inode *ino)
{
	unsigned int pos = 0;

	/* first two blocks are unused and superblock */
	pos += BLOCKSIZE * 2;

	/* inode and zone bitmaps */
	pos += BLOCKSIZE * sb.imap_blocks;
	pos += BLOCKSIZE * sb.zmap_blocks;

	/* and now the real inode (inode 0 has no entry in the table) */
	pos += (nr - 1) * sizeof(struct inode);

	READ(pos, sizeof(struct inode), ino);
}

/* returns physical zone number of the logical zone number. */
static unsigned short minix_get_zone(unsigned short logical,
				     const struct inode *ino)
{
	unsigned int addr;
	unsigned short rv;

	/* first seven zone entries of the inode are used to store the zone */
	if (logical < 7)
		return ino->zone[logical];
	logical -= 7;

	/* eight zone entry is used to store the zone number of a zone
	 * containing zone numbers */
	if (logical < ZONESIZE / sizeof(unsigned short)) {
		addr = ino->zone[7] * ZONESIZE + logical * 2;
		READ(addr, 2, &rv);
		return rv;
	}

	/* TODO: support this */
	panic("Requested file too big for me.");
}

int minix_read_data(const struct inode *ino, unsigned int addr,
		    void *buf, int count)
{
	unsigned short zone, blocksize;
	int rb = 0;
	unsigned int blockaddr;

	if (addr + count > ino->size)
		/* forbid reading after the file */
		count = ino->size - addr;
	if (count <= 0)
		/* nothing to read */
		return 0;

	while (count) {
		/* get zone number of the data we are going to read */
		zone = minix_get_zone(addr / ZONESIZE, ino);

		/* the size of the next block we read is the size of the area
		 * between our current start and the end of the zone, or our
		 * count if it is lower. */
		blocksize = ZONESIZE - (addr % ZONESIZE);
		blocksize = min(blocksize, count);

		/* calculate the addres of the zone + start of requestd area */
		blockaddr = (zone * ZONESIZE) + (addr % ZONESIZE);

		/* read blocksize bytes */
		READ(blockaddr, blocksize, buf + rb);

		addr  += blocksize;	/* start position of the block */
		rb    += blocksize;	/* number of already read bytes */
		count -= blocksize;	/* bytes we still have to read */
	}

	return rb;
}

static int minix_search(const char *file, const struct inode *ino, int links)
{
	struct inode ni;	/* next inode */
	struct dent dents[16];	/* 16 * sizeof(struct dent) = 512 = 1 block. */
	int i = 0, rv, c;
	int ctns;		/* chars to next slash */
	bool sf = false;
	struct dent *dent;

	if (!links--)
		return -ELOOP;

	/* skip first slash */
	file++;

	/* search next slash */
	ctns = strchr(file, '/') - file;
	if (ctns < 0) { /* no slash: we search the last file in the path */
		ctns = strlen(file);
		sf = true;
	} else if (ctns == 0)	/* '//' in filename. */
		return -EINVAL;

	do {
		rv = minix_read_data(ino, i, dents, sizeof(dents));
		i += rv;

		for (c = 0, dent = &dents[0];
		     c < rv;
		     c += sizeof(struct dent), dent++) {
			if (!strncmp(file, dent->name, ctns) &&
			    !dent->name[ctns]) {
				/* found matching file */
				if (sf)
					return dent->ino;
				minix_get_inode(dent->ino, &ni);
				if (!S_ISDIR(ni.mode))
					return -ENOTDIR;
				else if (!file[ctns+1])
					return dent->ino;
				else
					return minix_search(file + ctns, &ni,
							    links);
			}
		}
	} while (rv == sizeof(dents));

	/* the while loop should never end this way */
	return -ENOENT;
}

int minix_open(const char *file)
{
	struct inode ino;

	/* this must be here since minix_search does not work with '/' as
	 * parameter */
	if (file[0] == '/' && file[1] == 0)
		return 1;

	/* searching at files always starts at the root inode */
	minix_get_inode(1, &ino);

	return minix_search(file, &ino, MAX_LINKS);
}
