#include <stdbool.h>

#include "devices.h"
#include "lib.h"

/* these are our SPI spins. We have to implement software-SPI, since we only
 * have one SPI port. This one SPI port is already used by the enc28j60.
 *
 * There are two different ways to access MMC cards:
 *  1. The special MMC interface. According to my datasheet, it is faster, but
 *     in my opinion, it is really ugly (One pin is bidirectional to save pins,
 *     while an other pin is "reserverd for future use". What is even more
 *     horrible: The reserved pin should be pulled up with a TWO MEGAOHM
 *     resistor).
 *  2. The SPI interface. That's the one we use. I like it. */
#define CS	(1 << 21)	/* MMC pin 1 - Chip Select (active-low) */
#define DI	(1 << 20)	/* MMC pin 2 - Data In */
				/* MMC pin 3 - Ground */
				/* MMC pin 4 - Power Supply */
#define CLK	(1 << 14)	/* MMC pin 5 - Serial Clock (active-low) */
				/* MMC pin 6 - Ground */
#define DO	(1 << 19)	/* MMC pin 7 - Data Out */

/* MMC commands */
#define GO_IDLE_STATE		0	/* resets the MMC */
#define SEND_OP_COND		1	/* activates the MMC' initalization process */
#define READ_SINGLE_BLOCK	17	/* reads a single block */

#define BLOCKSIZE	512

/* number of times, something should be retried. */
#define TIMEOUT	2000

/* number of blocks in cache
 *
 * I made a benchmark with loading a big source tree from MMC via FTP and
 * measured the time needed for that with different CACHESIZE settings.
 * I got the following result:
 *  blocks	sec	diff
 *   2		49.389
 *   4		38.163	11.226
 *   6		32.272	 5.891
 *   8		28.137	 4.135
 *  10		25.348	 2.789
 *  12		24.380	 0.968
 *  14		23.600	 0.780
 *  16		22.853	 0.747
 *  18		22.701	 0.152 */
#define CACHESIZE	10
static struct {
	unsigned int blocknr;
	int value;

	enum {
		unused = 0,
		used   = 1
	} flags;

	char data[BLOCKSIZE];
} cacheent[CACHESIZE];

static void sspi_write(unsigned char byte)
{
	int i;

	for (i = 7; i >= 0; i--) {
		if (byte & (1 << i))
			IOSET = DI;
		else
			IOCLR = DI;

		IOCLR = CLK;
		IOSET = CLK;
	}

	IOSET = DI;
}

static unsigned char sspi_read()
{
	int i;
	unsigned char rv = 0;

	/* since we always keep DI on high when we do nothing with it, this
	 * function also sends the 0xff dummy byte. */

	for (i = 7; i >= 0; i--) {
		IOCLR = CLK;
		if (IOPIN & DO)
			rv |= (1 << i);
		IOSET = CLK;
	}

	return rv;
}

/* this function sends a command and leaves CS active after it. It returns the
 * answer */
static unsigned char mmc_command(int cmd, int argument)
{
	unsigned char instruction[6];
	unsigned char rv;
	int i;
	int timeout;

	/* the instructions have the following layout:
	 * 01, 6 bit command, 32 bit argument, 7 bit CRC, 1
	 *
	 * In SPI mode, the CRC needs not to be valid. Nevertheless, we have to
	 * send it when we sending the GO_IDLE_STATE command, since we are in
	 * MMC-Interface mode, when we do so (It's the command which switches to
	 * SPI). */
	instruction[0] = 0x40 | cmd;
	instruction[1] = (argument & 0xff000000) >> 24;
	instruction[2] = (argument & 0x00ff0000) >> 16;
	instruction[3] = (argument & 0x0000ff00) >> 8;
	instruction[4] = (argument & 0x000000ff);
	instruction[5] = (cmd == GO_IDLE_STATE) ? 0x95 : 0xff;

	/* send a dummy byte */
	IOSET = CS;
	sspi_write(0xff);

	/* send the instruction */
	IOCLR = CS;
	for (i = 0; i < 6; i++)
		sspi_write(instruction[i]);

	/* wait for a valid answer */
	for (rv = 0xff, timeout = TIMEOUT;
	     rv == 0xff && timeout;
	     rv = sspi_read(), timeout--) ;
	if (!timeout)
		panic("MultiMediaCard timed out.");

	return rv;
}

void mmc_init()
{
	int i;
	int timeout;

	/* define all output pins (= MMC' input pins) as output pins and set
	 * them to high */
	IODIR |= CS | DI | CLK;
	IOSET = CS | DI | CLK;

	/* send some SPI dummy bytes to the card */
	for (i = 0; i < 15; i++)
		sspi_write(0xff);

	/* GO_IDLE_STATE is used to reset the MMC and to tell it, that we hate
	 * the MMC interface and would like to use SPI */
	for (timeout = TIMEOUT; mmc_command(GO_IDLE_STATE, 0) != 1; timeout--) ;
	IOSET = CS;
	if (!timeout)
		panic("MultiMediaCard timed out.");

	/* Tell the card to initalize itself */
	for (timeout = TIMEOUT; mmc_command(SEND_OP_COND, 0) != 0; timeout--) ;
	IOSET = CS;
	if (!timeout)
		panic("MultiMediaCard timed out.");
}

/* reads a block */
static void mmc_read_block(unsigned int block, void *buf)
{
	int i;
	int timeout;

	mmc_command(READ_SINGLE_BLOCK, block * BLOCKSIZE);

	/* wait for start byte */
	for (timeout = TIMEOUT; sspi_read() != 0xfe; timeout--) ;
	if (!timeout)
		panic("MultiMediaCard timed out.");

	for (i = 0; i < BLOCKSIZE; i++)
		*((char *) buf++) = sspi_read();

	/* ignore 16 bit CRC sum */
	sspi_read();
	sspi_read();

	IOSET = CS;
}

/* reads a block, if possible from cache. */
static void mmc_read_cached_block(unsigned int block, void *buf)
{
	int i;
	int ent = 0;
	int lowest_v = INT_MAX, lowest_e = 0;
	bool found = 0;

	/* scan cache table. thereby, do:
	 * - check whether requested block has a valid entry
	 * - decrement value of each entry except of the requested one
	 * - increment value of requested block by number of cacheents + 1
	 * - find entry with lowest value. */
	for (i = 0; i < CACHESIZE; i++) {
		if (cacheent[i].blocknr == block && cacheent[i].flags == used) {
			found = 1;
			ent = i;
			if (cacheent[i].value < INT_MAX - CACHESIZE - 1)
				cacheent[i].value += CACHESIZE + 1;
		} else if (cacheent[i].flags == used)
			if (cacheent[i].value > INT_MIN + 1)
				cacheent[i].value--;

		if (cacheent[i].value < lowest_v) {
			lowest_v = cacheent[i].value;
			lowest_e = i;
		}
	}

	if (found) {
		/* block was found. read it from cache */
		memcpy(buf, cacheent[ent].data, BLOCKSIZE);
		return;
	}

	/* block not found. read it from MMC and cache it if possible. */

	mmc_read_block(block, buf);

	if (lowest_v > 0)
		/* value of lowest-valued cache entry too high. we do not
		 * replace it. */
		return;

	cacheent[lowest_e].blocknr = block;
	cacheent[lowest_e].value = CACHESIZE + 1;
	cacheent[lowest_e].flags = used;
	memcpy(cacheent[lowest_e].data, buf, 512);
}

/* reads a part of a block. returns the number of bytes read */
static unsigned int mmc_do_read(unsigned int addr, unsigned int count,
				void *buf)
{
	char tmpbuf[BLOCKSIZE];
	unsigned int i, j;

	mmc_read_cached_block(addr / BLOCKSIZE, tmpbuf);

	for (i = addr % BLOCKSIZE, j = 0;
	     j < count;
	     i++, j++)
		((char *) buf)[j] = tmpbuf[i];

	return j;
}

void mmc_read(unsigned int addr, unsigned int count, void *buf)
{
	unsigned int r, cc;
	unsigned int j = 0, pos = addr;

	while (j < count) {
		/* Three things must be considered for setting cc:
		 * 1. current addr might be behind the beginning of a block
		 * 2. current addr might be at the beginning of a block
		 * 3. current addr might be in the last requested block */
		cc = BLOCKSIZE - pos % BLOCKSIZE;	/* 1 */
		if (!cc)				/* 2 */
			cc = BLOCKSIZE;
		if (pos + cc > addr + count)		/* 3 */
			cc = addr + count - pos;

		r = mmc_do_read(pos, cc, buf + j);

		pos += r;
		j   += r;
	}
}
