#ifndef NETWORK_H
#define NETWORK_H

#include "lib.h"

/* IP address */
#define IP_ADDR	192,168,9,41
extern const unsigned char ip_addr[4];

/* UDP console configuration */
#define UC_PORT		0xa661				/* 24998 */
#define UC_DSTIP	192,168,9,42
#define UC_DSTMAC	0x00,0x86,0x87,0xEC,0x72,0x3A

/* Time to Live for IP-packets we send */
#define TTL	64

/* these are in network byte order. so we save useless conversions */
#define ETH_IP	0x0008
#define ETH_ARP	0x0608

#define IP_ICMP	0x01
#define IP_TCP	0x06
#define IP_UDP	0x11

/* just used for computing UDP and TCP checksum */
struct pseudo_header {
	unsigned char sip[4];
	unsigned char dip[4];
	unsigned char zero, ptcl;
	unsigned short length;
} __attribute__ ((packed));

extern void arp_handle(void *pkg, const unsigned char *src);
extern void ip_handle(void *pkg, const unsigned char *src);

extern void ip_transmit(const unsigned char *dstmac, const unsigned char *dstip,
			unsigned short type, const void *pkg, unsigned int len);

extern void udp_transmit(const unsigned char *dstmac, const unsigned char *dstip,
			 unsigned short dstport, unsigned short srcport,
			 const void *pkg, unsigned int len);

extern void udpcons_putchar(char c);

extern void icmp_handle(void *pkg, unsigned short len,
			const unsigned char *srcmac,
			const unsigned char *srcip);
extern void tcp_handle(void *pkg, unsigned short len,
		       const unsigned char *srcmac, const unsigned char *srcip);

struct listen;

struct tcp_context {
	/* source addresses */
	unsigned char smac[6];
	unsigned char sip[4];

	/* ports are stored in network byte order, cause they need no real
	 * processing. */
	unsigned short sport;
	unsigned short dport;

	/* stored in host byte order */
	unsigned int nextseq;
	unsigned int nextack;
	unsigned int lastseq;
	unsigned int lastack;
	unsigned int nextplus;
	int window;

	/* current connection state */
	enum {
		closed,		/* no connection */
		connecting,	/* SYN received */
		established,	/* ACK received */
		transmitting,	/* transmitting data */
		closing		/* FIN|ACK sent/received */
	} state;

	/* transmitter callback. set with tcp_start_transmit() */
	void (*send_packet)(struct tcp_context *ctx);

	/* place for handler to store data */
	void *userdata;

	/* pointer to server app info */
	const struct listen *listen;

	/* pointer to next context (this is a linked list) */
	struct tcp_context *next;
};

extern void tcp_transmit_pkg(struct tcp_context *ctx, const void *pkg, int len);

/* just an helper makro */
#define tcp_transmit_str(ctx, str) \
	tcp_transmit_pkg(ctx, str, strlen(str))

/* this function is used to start a packet transmission.
 * send_packet() is a function which is called everytime, a packet is ready to
 * be sent. It should either call tcp_transmit_pkg(), tcp_stop_transmit(), or
 * tcp_disconnect() */
extern void tcp_start_transmit(struct tcp_context *ctx,
			       void (*send_packet)(struct tcp_context *ctx));
extern void tcp_stop_transmit(struct tcp_context *ctx);

/* closes the connection. It does NOT call the close callback. So any userdata
 * in the context must be freed. */
extern void tcp_disconnect(struct tcp_context *ctx);

extern void http_connect(struct tcp_context *);
extern void http_push(struct tcp_context *, void *, unsigned short);
extern void http_close(const struct tcp_context *);

extern void ftp_connect(struct tcp_context *);
extern void ftp_push(struct tcp_context *, void *, unsigned short);
extern void ftp_close(const struct tcp_context *);

extern void ftpdata_connect(struct tcp_context *);
extern void ftpdata_push(struct tcp_context *, void *, unsigned short);
extern void ftpdata_close(const struct tcp_context *);

#endif
