#include "devices.h"
#include "lib.h"
#include "network.h"

/* TCP is defined in RFC 793. */

/*			       EXAMPLE CAPTURE
 *
 * ac	fl	sport	dport	seq		ack		datalen
 * send	syn	sp	dp	s0		0
 * recv	syn,ack	dp	sp	s1		s0+1
 * send	ack	sp	dp	s0+1		s1+1
 * recv	psh,ack	dp	sp	s1+1		s0+1		6
 * send	ack	sp	dp	s0+1		s1+1+6
 * send	psh,ack	sp	dp	s0+1		s1+1+6		3
 * recv	ack	dp	sp	s1+1+6		s0+1+3
 * recv	psh,ack	dp	sp	s1+1+6		s0+1+3		3
 * send	ack	sp	dp	s0+1+3		s1+1+6+3
 * send	psh,ack	sp	dp	s0+1+3		s1+1+6+3	5
 * recv	ack	dp	sp	s1+1+6+3	s0+1+3+5
 * recv	fin,ack	dp	sp	s1+1+6+3	s0+1+3+5
 * send	fin,ack	sp	dp	s0+1+3+5	s1+1+6+3+1
 * recv	ack	dp	sp	s1+1+6+3+1	s0+1+3+5+1
 *
 * 			      PROTOCOL OVERVIEW
 *
 * acking:
 *
 * - except of ACK-only, everything has to be ACKed
 * - even PSH include ACK
 *
 * - plus = (depends on what is ACKed)
 * 	PSH: last datalen
 * 	FIN/SYN: 1
 * 	ACK-only: 0
 *
 * - send ACK|...	with seq=lastack, ack=lastseq+plus
 *
 * - seq and ack numbers are updated when packet is received. not when send
 *
 * window:
 *
 * - each packet contains a window size, specifying how big the packet buffer of
 *   the socket is
 *
 * connection startup:
 *
 * - send SYN		with seq=s0,	ack=0
 * - recv SYN|ACK	with seq=s1,	ack=s0+1
 * - send ACK		with seq=s0+1,	ack=s1+1
 *
 * (seq and ack are as defined above ("acking"), except of that s1 does not need
 * to be 0.)
 *
 * pushing:
 *
 * - send PSH|ACK
 * - recv ACK
 *
 * (both with seq and ack as defined above ("acking"))
 *
 * - When there is enough data to send, there is the possibility to do more than
 *   one PSH|ACK without waiting for their ACK, till the opposite window is full.
 *   In this case, only the last PSH|ACK needs to be ACKed. The seq and ack
 *   numbers are, as they would be, when every PSH|ACK would be acked.
 *
 * connection closing:
 *
 * - send FIN|ACK
 * - recv FIN|ACK
 * - send ACK
 *
 * (both with seq and ack as defined above ("acking"))
 *
 * 			 IMPLEMENTATION INTRODUCTION
 *
 * per-socket structure:
 *
 * - standard stuff (mac addr, ip addr, src/dst ports, list pointers, callbacks,
 *   ...)
 * - next seq to be sent
 * - next ack to be sent
 * - last sent seq
 * - last sent ack
 * - nextplus
 * - opposite window size
 * - connection state (CONNECTING, ESTABLISHED, TRANSFERRING, CLOSING)
 *
 * calc_plus:
 *
 * - rv = datalen
 * - fl & SYN -> rv++
 * - fl & FIN -> rv++
 * - return rv
 *
 * transmit:
 *
 * - fill headers etc. (with nextack, nextseq, ...)
 *  - window size is:
 *   - rv = receive buffer size / number of sockets
 *   - if (rv < 1518)
 *    - rv = min(1518, receive buffer size)
 *
 * - transmit ip packet
 *
 * - nextplus = calc_plus
 * - lastseq = nextseq
 * - lastack = nextack
 * - nextseq = lastseq + nextplus
 * - window = window - (tcplen+iplen+ethlen)
 *
 * push:
 *
 * - transmit PSH|ACK
 *
 * close:
 *
 * - state = CLOSING
 * - transmit FIN|ACK
 *
 * continue_transfer:
 *
 * - while window > 1518
 * - call push_possible callback
 *
 * start_transmit:
 *
 * - state = TRANSMITTING
 * - continue_transfer
 *
 * stop_transmit:
 *
 * - state = ESTABLISHED
 *
 * packet receiption:
 *
 * - fl == SYN:
 *  - search server application. if not found:
 *   - transmit RST|ACK
 *   - return
 *  - record new socket:
 *   - n.nextseq = 0
 *   - n.nextack = seq+1
 *   - n.window = window
 *  - transmit SYN|ACK
 *  - state = CONNECTING
 *  - return
 *
 * - if !(fl & ACK), discard package
 *
 * - search socket by ip and source port
 *
 * - if seq != n.lastack or ack != n.lastseq+n.nextplus, discard package
 *
 * - n.window = window
 *
 * - n.nextseq = ack
 * - n.nextack = seq+calc_plus
 *
 * - fl == ACK:
 *  - if state = CONNECTING
 *   - state = ESTABLISHED
 *   - call established callback
 *
 * - if state = TRANSFERRING
 *  - continue_transfer
 *
 * - fl & PSH
 *  - forward package to server application
 *  - transmit ACK
 *
 * - fl & FIN
 *  - if state = CLOSING
 *   - transmit ACK
 *  - else
 *   - transmit FIN|ACK
 *   - call close handler
 *   - free socket data */

#define FIN	(1 << 0)
#define SYN	(1 << 1)
#define RST	(1 << 2)
#define PUSH	(1 << 3)
#define ACK	(1 << 4)
#define URG	(1 << 5)

struct tcp_header {
	unsigned short sport;	/* source port */
	unsigned short dport;	/* destination port */
	unsigned int seq;	/* sequence number */
	unsigned int ack;	/* acknowledgement number */
	unsigned char hdrlen;	/* upper: header length in words
				 * lower: 0 */
	unsigned char flags;
	unsigned short window;	/* receive window size */
	unsigned short cs;	/* checksum */
	unsigned short urp;	/* urgent pointer */
} __attribute__((packed));

static struct tcp_context *ctx_head;
static int ctxnum;	/* number of contexts */

static const struct listen {
	/* port number in network byte order */
	unsigned short port;

	/* callback for new connection. It is called after the
	 * three-way-handshake is done when the connection is fully established.
	 * You may use it to setup userdata, start transmissions or to close the
	 * connection. */
	void (*connect)(struct tcp_context *);

	/* callback for new data. It is called after new data arrived and was
	 * ACKed.
	 * You may use it to start transmissions or to close the connection. */
	void (*push)(struct tcp_context *, void *, unsigned short);

	/* callback for disconnect by other side.
	 * It is called, when the FIN|ACK was received when the connection is
	 * not fully closed.
	 * It is only called when the disconnect attempt is done by the other
	 * side. Not, if this side attempts to disconnect.
	 * When disconnecting or getting disconnected, the userdata should be
	 * freed. The TCP implementation itself does NOT free any userdata.
	 * You may use it to free userdata (What you also should do, when you
	 * initiate the disconnect). */
	void (*close)(const struct tcp_context *);
} listen[] = {
	{ 0x1400, ftpdata_connect, ftpdata_push, ftpdata_close },
	{ 0x1500, ftp_connect,     ftp_push,     ftp_close     },
	{ 0x5000, http_connect,    http_push,    http_close    }
};

/* transmit raw TCP package */
static void tcp_transmit(const unsigned char *dmac, const unsigned char *dip,
			 unsigned short dport, unsigned short sport,
			 unsigned int seq, unsigned int ack,
			 unsigned char flags, const void *pkg, int pkglen,
			 int wsize)
{
	int rpllen = sizeof(struct pseudo_header) + sizeof(struct tcp_header) +
		pkglen;
	struct pseudo_header *pseudo_header;
	struct tcp_header *tcp_header;
	char rpl[rpllen];
	pseudo_header = (void *) rpl;
	tcp_header = ((void *) rpl) + sizeof(struct pseudo_header);

	/* fill all headers */
	memcpy(pseudo_header->sip, ip_addr, 4);
	memcpy(pseudo_header->dip, dip, 4);
	pseudo_header->zero = 0;
	pseudo_header->ptcl = IP_TCP;
	pseudo_header->length = htons(rpllen - sizeof(struct pseudo_header));
	tcp_header->sport = sport;
	tcp_header->dport = dport;
	tcp_header->seq = htonl(seq);
	tcp_header->ack = htonl(ack);
	tcp_header->hdrlen = (sizeof(struct tcp_header) / 4) << 4;
	tcp_header->flags = flags;
	tcp_header->window = htons(wsize);
	tcp_header->cs = 0;	/* computed later */
	tcp_header->urp = 0;

	/* copy data */
	memcpy(&rpl[rpllen - pkglen], pkg, pkglen);

	/* compute checksum */
	tcp_header->cs = create_checksum(rpl, rpllen);

	/* transmit */
	ip_transmit(dmac, dip, IP_TCP, rpl + sizeof(struct pseudo_header),
		    rpllen - sizeof(struct pseudo_header));
}

/* allocate new tcp context */
static struct tcp_context *tcp_new_ctx()
{
	struct tcp_context *tcpctx;

	ctxnum++;

	if (!ctx_head) {
		/* creating first tcpctx */
		ctx_head = malloc(sizeof(struct tcp_context));
		ctx_head->next = 0;
		return ctx_head;
	}

	/* find last tcpctx in queue */
	tcpctx = ctx_head;
	while (tcpctx->next)
		tcpctx = tcpctx->next;

	/* create entry */
	tcpctx->next = malloc(sizeof(struct tcp_context));
	tcpctx->next->next = 0;
	return tcpctx->next;
}

/* search listener information */
static const struct listen *tcp_search_listen(unsigned short port)
{
	unsigned int i;

	for (i = 0; i < ARRAY_SIZE(listen); i++)
		if (listen[i].port == port)
			return &listen[i];

	return 0;
}

/* search belonging tcp context */
static struct tcp_context *tcp_search_ctx(const struct tcp_header *hdr,
					  const unsigned char *sip)
{
	struct tcp_context *ctx = ctx_head;

	/* due to performance issues, we only check ip and port. */

	while (ctx) {
		if (ctx->sport == hdr->sport && !memcmp(ctx->sip, sip, 4))
			return ctx;

		ctx = ctx->next;
	}

	return 0;	/* not found */
}

/* unlinks and frees tcp context */
static void tcp_delete_ctx(const struct tcp_context *ctx)
{
	struct tcp_context *cur = ctx_head, *next = ctx->next;

	ctxnum--;

	/* free context */
	free(ctx);

	/* search context in list and unlink it */
	if (ctx_head == ctx) {
		ctx_head = next;
		return;
	}
	do {
		if (cur->next == ctx) {
			cur->next = next;
			return;
		}
	} while ((cur = cur->next) != 0);

	/* if we are here, the context was not linked...
	 * That's a case which should never happen (compiler bug, dirty code, or
	 * something like that) */
}

/* calculates the seq/ack increment */
static unsigned int calc_plus(unsigned char flags, int pkglen)
{
	unsigned int rv = 0;

	if (flags & PUSH)
		rv += pkglen;
	if (flags & SYN)
		rv++;
	if (flags & FIN)
		rv++;

	return rv;
}

/* returns our current window size per socket */
static int tcp_own_window()
{
	int wsize;
	int recvbufsiz;

	recvbufsiz = enc28j60_get_recv_window();
	wsize = recvbufsiz / ctxnum;

	if (wsize < 1518)
		return min(1518, recvbufsiz);
	else
		return min(2048, wsize);
}

/* transmit TCP package */
static void tcp_transmit_ctx(struct tcp_context *ctx, unsigned char flags,
			     const void *pkg, int pkglen)
{
	int wsize = tcp_own_window();

	tcp_transmit(ctx->smac, ctx->sip, ctx->sport, ctx->dport, ctx->nextseq,
		     ctx->nextack, flags, pkg, pkglen, wsize);

	ctx->nextplus = calc_plus(flags, pkglen);
	ctx->lastseq = ctx->nextseq;
	ctx->lastack = ctx->nextack;
	ctx->nextseq = ctx->lastseq + ctx->nextplus;
	ctx->window = ctx->window - pkglen - sizeof(struct tcp_header) -
		sizeof(struct pseudo_header) - 18;
}

static void tcp_continue_transmit(struct tcp_context *ctx)
{
	int maxwin = tcp_own_window();

	while (ctx->window >= 1518 && ctx->state == transmitting &&
	       maxwin > 0) {
		ctx->send_packet(ctx);
		maxwin -= 128;
	}
}

void tcp_handle(void *pkg, unsigned short len, const unsigned char *srcmac,
		const unsigned char *srcip)
{
	const struct listen *server_app;
	struct tcp_header *tcp_header = pkg;
	struct tcp_context *ctx;
	int pkglen;

	unsigned short hdrlen = ((tcp_header->hdrlen & 0xf0) >> 4) * 4;
	pkg += hdrlen;

	pkglen = len - hdrlen;

	/* do some endianess conversions */
	tcp_header->seq = ntohl(tcp_header->seq);
	tcp_header->ack = ntohl(tcp_header->ack);
	tcp_header->window = ntohs(tcp_header->window);

	if (tcp_header->flags == SYN) {
		server_app = tcp_search_listen(tcp_header->dport);
		if (!server_app) {
			tcp_transmit(srcmac, srcip, tcp_header->sport,
				     tcp_header->dport, 0, tcp_header->seq + 1,
				     RST|ACK, 0, 0, 64);
			return;
		}

		ctx = tcp_new_ctx();
		ctx->listen = server_app;
		ctx->nextseq = 0;
		ctx->nextack = tcp_header->seq + 1;
		ctx->window = tcp_header->window;
		ctx->state = connecting;
		memcpy(ctx->smac, srcmac, 6);
		memcpy(ctx->sip, srcip, 4);
		ctx->sport = tcp_header->sport;
		ctx->dport = tcp_header->dport;
		tcp_transmit_ctx(ctx, SYN|ACK, 0, 0);
		return;
	}

	ctx = tcp_search_ctx(tcp_header, srcip);
	if (!ctx)
		return;

	if (tcp_header->seq != ctx->lastack)
		return;

	if (tcp_header->flags & RST) {
		tcp_transmit_ctx(ctx, RST|ACK, 0, 0);
		ctx->listen->close(ctx);
		tcp_delete_ctx(ctx);
		return;
	}

	if (!(tcp_header->flags & ACK))
		return;

	if (tcp_header->ack != ctx->lastseq + ctx->nextplus)
		return;

	ctx->window = tcp_header->window;

	ctx->nextseq = tcp_header->ack;
	ctx->nextack = tcp_header->seq + calc_plus(tcp_header->flags, pkglen);

	if (tcp_header->flags == ACK)
		if (ctx->state == connecting) {
			ctx->state = established;
			ctx->listen->connect(ctx);
		}

	if (ctx->state == transmitting)
		tcp_continue_transmit(ctx);

	if (tcp_header->flags & PUSH) {
		tcp_transmit_ctx(ctx, ACK, 0, 0);
		ctx->listen->push(ctx, pkg, pkglen);
	}

	if (tcp_header->flags & FIN) {
		if (ctx->state == closing)
			tcp_transmit_ctx(ctx, ACK, 0, 0);
		else {
			tcp_transmit_ctx(ctx, FIN|ACK, 0, 0);
			ctx->listen->close(ctx);
		}

		tcp_delete_ctx(ctx);
	}
}

void tcp_transmit_pkg(struct tcp_context *ctx, const void *pkg, int len)
{
	tcp_transmit_ctx(ctx, PUSH|ACK, pkg, len);
}

void tcp_start_transmit(struct tcp_context *ctx,
			void (*send_packet)(struct tcp_context *ctx))
{
	ctx->state = transmitting;
	ctx->send_packet = send_packet;

	tcp_continue_transmit(ctx);
}

void tcp_stop_transmit(struct tcp_context *ctx)
{
	ctx->state = established;
}

void tcp_disconnect(struct tcp_context *ctx)
{
	ctx->state = closing;
	tcp_transmit_ctx(ctx, FIN|ACK, 0, 0);
}
